﻿/* TODO:
 * - Lose the TCP connection when fatal errors occur... follow the semantics of Twisted's implementation.
 *   we currently lose the connection in certain cases, but not others. It's all very ambiguous.
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Net.Sockets;
using System.Threading;
using System.Runtime.CompilerServices;
using System.Windows.Forms;
using Timer=System.Threading.Timer;

// Configure logging for this assembly using the AMPSharp.dll.log4net file
[assembly: log4net.Config.XmlConfigurator(ConfigFileExtension="log4net", Watch=true)]

namespace AMP
{
    // The return value of AMP_Responder must be in the form of Object since we allow you
    // to return an AMP_Msg (if responding synchronously) or a DelegateBox (if you need to
    // respond asyncronously).
    public delegate Object AMP_Responder(AMP_Msg msg, Object state);

    // Ideally we would have full-featured Deferreds in C#, but we don't right now, so
    // we use a thing called DelegateBox. You can return one of these from your
    // AMP_Responder delegate and AMP will then attach a single callback to it.
    //
    // Then later you can call DelegateBox.callback(AMP_Msg msg) to
    // respond to the original AMP command.
    //
    // Or, to fail the AMP request, simply pass an Exception to DelegateBox.callback
    public delegate void DelegateBoxCallback(Object amp_msg_or_exception);

    /// <summary>
    /// Used like a Deferred. Except this is nothing like real Deferreds.
    /// This is the most brain-dead, simple abstraction for a deferred callback possible,
    /// and it's completely specific to our AMP implementation..that's intentional.. anything
    /// I develop in the future for more general use will be called 'Deferred' and have all
    /// the bells and whistles of real Deferreds.
    /// </summary>
    public class DelegateBox
    {
        public DelegateBoxCallback _theCallback;
        public Object _savedResult;
        public void setCallback(DelegateBoxCallback callback)
        {
            if (_savedResult != null)
            {
                // run callback synchronously
                callback(_savedResult);
            }
            else
            {
                _theCallback = callback;
            }
        }
        public void callback(Object result)
        {
            if (_theCallback == null)
            {
                _savedResult = result;
            } 
            else
            {
                _theCallback(result);
            }
        }
    }

    public interface IAMPTransport
    {
        void RegisterProtocol(IAMPProtocol proto);
        void Start();
        void Write(byte[] bytes);
    }

    public class AMPSocketTransport: IAMPTransport
    {
		private static readonly log4net.ILog log = log4net.LogManager.GetLogger("AMP");
        private const int readBufSize = 32768;
        private readonly byte[] readBuf = new byte[readBufSize];

        public Socket socket;
        private IAMPProtocol proto;
        public AMPSocketTransport(Socket socket)
        {
            this.socket = socket;
        }

        public Socket Socket
        {
            get { return socket; }
        }

        public void Start()
        {
            socket.BeginReceive(readBuf, 0, readBuf.Length, 0, new AsyncCallback(async_dataReceived), socket);
        }

        public void Write(byte[] bytes)
        {
            try
            {
                socket.BeginSend(bytes, 0, bytes.Length, 0, new AsyncCallback(async_dataSent), socket);
            }
            catch (Exception e)
            {
                _handleFatalNetworkError(e);
            }
        }

        public void async_dataSent(IAsyncResult ar)
        {
            int bytesSent;
            var s = (Socket)ar.AsyncState;
            try
            {
                // XXX TODO can we have a partial send here???
                // If so we need to write a test case and implement handling code.
                bytesSent = s.EndSend(ar);
            }
            catch (Exception e)
            {
                log.Error("Error sending to Server:", e);
                _handleFatalNetworkError(e);
                return;
            }
            log.Debug("Sent bytes to peer: " + bytesSent);
        }

        public void RegisterProtocol(IAMPProtocol proto)
        {
            this.proto = proto;
        }

        public void async_dataReceived(IAsyncResult ar)
        {
            // XXX TODO I'm not 100% sure but it might be possible that this callback gets called
            // even after this socket and protocolParser has been discarded. Which means we will
            // have failed any outstanding callRemote's that these bytes might be related too..
            // We should ignore these bytes in that case.
            //
            // UPDATE: Yes, it's definately possible.

            int bytesRead;
            var s = (Socket)ar.AsyncState;

            // XXX TODO need to call proto.ConnectionLost() here?
            if (!s.Connected) return; // Don't need to go any further if we're disconnected

            try
            {
                bytesRead = s.EndReceive(ar);
            }
            catch (Exception e)
            {
                log.Error("Error reading from socket:", e);
                _handleFatalNetworkError(e);
                return;
            }
            log.Debug("Read bytes from peer: " + bytesRead);
            if (bytesRead == 0)
            {
                log.Debug("Connection to peer closed cleanly.");
                proto.ConnectionLost(new SocketClosedException());

            }
            else
            {
                proto.DataReceived(readBuf, bytesRead);

                try
                {
                    s.BeginReceive(readBuf, 0, readBuf.Length, 0, new AsyncCallback(async_dataReceived), s);
                }
                catch (Exception e)
                {
                    log.Error("Error calling BeginReceive:", e);
                    _handleFatalNetworkError(e);
                }

            }
        }

        public void _handleFatalNetworkError(Exception err)
        {
            // XXX TODO we need to be double-plus sure that we 
            // don't call ConnectionLost more than once.
            proto.ConnectionLost(err);
        }

    }

    public interface IAMPProtocol
    {
        AMP_Msg CallRemote(AMP_Command cmd, AMP_Msg msg);
        IAsyncResult BeginCallRemote(AMP_Command cmd, AMP_Msg msg, AsyncCallback callback, Object cbState);
        AMP_Msg EndCallRemote(IAsyncResult ar);
        void RegisterResponder(AMP_Command cmd, AMP_Responder responder, Object callbackState);
        void RegisterResponder(AMP_Command cmd, AMP_Responder responder);
        void RegisterResponder(AMP_Command cmd, AMP_Responder responder, Object callbackState, SynchronizationContext callbackContext);
        void RegisterResponder(AMP_Command cmd, AMP_Responder responder, SynchronizationContext callbackContext);
        void RegisterResponder(AMP_Command cmd, AMP_Responder responder, Object callbackState, Control callbackControl);
        void RegisterResponder(AMP_Command cmd, AMP_Responder responder, Control callbackControl);
        void DataReceived(byte[] bytes, int len);
        void ConnectionLost(Exception reason);
    }

    /// <summary>
    /// AMP protocol-parsing and state-keeping class.
    /// Instantiate it with an IAMPTransport that is already be connected to the remote peer.
    /// Be sure to call StartReading() on the IAMPTransport
    /// Provides CallRemote() and BeginCallRemote(), for making remote AMP calls.
    /// Provides RegisterResponder() for registering delegates that will handle
    /// the AMP commands that we support on this side of the connection. Responders
    /// may be either synchrounous or asynchronous.
    /// </summary>
    public class AMP_Protocol: IAMPProtocol
    {
		private static readonly log4net.ILog log = log4net.LogManager.GetLogger("AMP");
        public double callRemoteTimeout = 60.0;

        // TODO use real constants for this
        //public const int KEY_LEN_READ = 0;
        //public const int KEY_DATA_READ = 1;
        //public const int VAL_LEN_READ = 2;
        //public const int VAL_DATA_READ = 3;
        //public const int FULL_MESSAGE_FINISHED = 4;

        private enum states
        {
            KEY_LEN_READ,
            KEY_DATA_READ,
            VAL_LEN_READ,
            VAL_DATA_READ,
            FULL_MESSAGE_FINISHED
        };

        private states state;
        public ArrayList key_len;
        public ArrayList key_data;
        public ArrayList val_len;
        public int       val_len_int;
        public byte[] val_data;

        public AMP_Msg_Raw amp_msg_raw;

        public int lastAskKey;
        public Dictionary<int, Ask_Info> askInfoDict;

        // a place to stuff Timer object so that they don't get garbage collected before we want them to be
        internal Dictionary<int, Timer> tempTimers = new Dictionary<int, Timer>();
        internal int lastTempTimerKey;

        private readonly Dictionary<string, responderStruct> registeredCommands = new Dictionary<string, responderStruct>();

        public IAMPTransport transport;

        public AMP_Protocol()
        {
            state = states.KEY_LEN_READ;
            key_len = new ArrayList();
            key_data = new ArrayList();
            val_len = new ArrayList();
            val_data = new byte[0];

            amp_msg_raw = new AMP_Msg_Raw();

            askInfoDict = new Dictionary<int, Ask_Info>();

        }

        public AMP_Protocol(IAMPTransport transport)
            : this()
        {
            this.transport = transport;
            transport.RegisterProtocol(this);
        }

        // Hold a reference to an AMP_Responder, the AMP_Command for this responder and the responder state object
        struct responderStruct
        {
            public AMP_Responder responder;
            public AMP_Command command;
            public Object state; // may be null
            public SynchronizationContext callbackContext; // may be null
            public Control callbackControl; // may be null
        }

        public class CommandAlreadyRegistered: Exception { };

        public void RegisterResponder(AMP_Command cmd, AMP_Responder responder)
        {
            _RegisterResponder(cmd, responder, null, null, null);
        }

        public void RegisterResponder(AMP_Command cmd, AMP_Responder responder, object callbackState, SynchronizationContext callbackContext)
        {
            _RegisterResponder(cmd, responder, callbackState, callbackContext, null);
        }

        public void RegisterResponder(AMP_Command cmd, AMP_Responder responder, SynchronizationContext callbackContext)
        {
            _RegisterResponder(cmd, responder, null, callbackContext, null);
        }

        public void RegisterResponder(AMP_Command cmd, AMP_Responder responder, object callbackState, Control callbackControl)
        {
            _RegisterResponder(cmd, responder, callbackState, null, callbackControl);
        }

        public void RegisterResponder(AMP_Command cmd, AMP_Responder responder, Control callbackControl)
        {
            _RegisterResponder(cmd, responder, null, null, callbackControl);
        }
        public void RegisterResponder(AMP_Command cmd, AMP_Responder responder, Object callbackState)
        {
            _RegisterResponder(cmd, responder, callbackState, null, null);
        }
        private void _RegisterResponder(AMP_Command cmd, AMP_Responder responder, Object callbackState,
            SynchronizationContext callbackContext, Control callbackControl)
        {
            if (registeredCommands.ContainsKey(cmd.name))
            {
                throw new CommandAlreadyRegistered();
            }
            var respRef = new responderStruct {responder = responder,
                command = cmd, state = callbackState, callbackContext = callbackContext,
                callbackControl = callbackControl};
            registeredCommands.Add(cmd.name, respRef);
        }

        public virtual void ConnectionLost(Exception reason)
        {
            foreach (Ask_Info info in askInfoDict.Values)
            {
                info.ar._error = reason;
                try
                {
                    info.ar.doCallback();
                }
                catch (Exception e)
                {
                    log.Error("Exception in user-defined AMP callback while failing outstanding requests:", e);
                }
            }

            // dont bother to do this if we are gonna be closing the socket associated with
            // this protocol.. 
            askInfoDict.Clear();
        }
        //public virtual void handleFatalNetworkError(Exception err)
        //{
        //}

        public void DataReceived(byte[] buf)
        {
            DataReceived(buf, buf.Length);
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        public void DataReceived(byte[] buf, int len)
        {
            int idx = 0;

            while (idx < len)
            {
                int bytes_left = len - idx;
                int bytes_needed;
                switch (state)
                {
                    case states.KEY_LEN_READ:
                        if (key_len.Count == 0 && buf[idx] != 0)
                        {
                            // The 1st byte of the key length must always be \x00
                            throw new AMPProtocolError();
                        }
                        key_len.Add(buf[idx]);
                        idx++;
                        if (key_len.Count == 2) {
                            // got the 2 bytes for a key. transition state.
                            // if these 2 bytes are both null, then that should signal the
                            // end of a full AMP message.
                            if ((byte)key_len[0] == 0 && (byte)key_len[1] == 0) {
                                if (amp_msg_raw.Count == 0) {
                                    // we haven't received any key/values yet so there shouldn't be a 
                                    // double-NULL here
                                    throw new AMPProtocolError();
                                }
                                // We got the empty-key terminator on a full, valid message, yay!
                                key_len.Clear(); // forget the 2 null bytes
                                processFullMessage(amp_msg_raw);
                                state = states.FULL_MESSAGE_FINISHED; // transition
                            } else {
                                state = states.KEY_DATA_READ; // transition
                            }
                        } // else state stays the same since we've only read 1 byte of the 2-byte key length
                        break;

                    case states.KEY_DATA_READ:
                        int key_len_int = (byte)key_len[1];
                        // and thus are only represented by the 2nd byte
                        bytes_needed = key_len_int - key_data.Count;
                        byte[] tmp;
                        if (bytes_left < bytes_needed) {
                            tmp = new byte[bytes_left];
                            Array.Copy(buf, idx, tmp, 0, bytes_left);
                            key_data.AddRange(tmp);
                            //key_data.append(buf+idx, bytes_left);
                            idx += bytes_left;
                            // state stays the same
                        } else if (bytes_left >= bytes_needed) {
                            tmp = new byte[bytes_needed];
                            Array.Copy(buf, idx, tmp, 0, bytes_needed);
                            key_data.AddRange(tmp);
                            //key_data.append(buf+idx, bytes_needed);
                            idx += bytes_needed;
                            state = states.VAL_LEN_READ; // transition
                        }

                        break;
                    case states.VAL_LEN_READ:
                        val_len.Add(buf[idx]);
                        idx++;
                        if (val_len.Count == 2) {
                            // got the 2 bytes for a value length.
                            var tmp_bytes = new byte[2];
                            if (BitConverter.IsLittleEndian)
                            {
                                // reverse bytes before casting because this is a little-endian computer, but the bytes
                                // we read are in network byte order (big-endian)
                                tmp_bytes[0] = (byte)val_len[1];
                                tmp_bytes[1] = (byte)val_len[0];
                            }
                            else
                            {
                                tmp_bytes[0] = (byte)val_len[0];
                                tmp_bytes[1] = (byte)val_len[1];
                            }
                            val_len_int = BitConverter.ToUInt16(tmp_bytes, 0);

                            state = states.VAL_DATA_READ; // transition
                        }
                        // else the state stays the same since we've only read 1 byte of the 2-byte val length
                        break;
                    case states.VAL_DATA_READ:
                        //bytes_needed = val_len_int - val_data.Count;
                        bytes_needed = val_len_int - val_data.Length;
                        if (bytes_left < bytes_needed) {
                            var tmp_bytes = new byte[val_data.Length+bytes_left];
                            Array.Copy(val_data, tmp_bytes, val_data.Length);
                            Array.Copy(buf, idx, tmp_bytes, val_data.Length, bytes_left);
                            val_data = tmp_bytes;
                            idx += bytes_left;
                            // state stays the same
                        } else if (bytes_left >= bytes_needed) {
                            var tmp_bytes = new byte[val_data.Length+bytes_needed];
                            Array.Copy(val_data, tmp_bytes, val_data.Length);
                            Array.Copy(buf, idx, tmp_bytes, val_data.Length, bytes_needed);
                            val_data = tmp_bytes;
                            idx += bytes_needed;
                            state = states.KEY_LEN_READ; // transition
                            processCurrentKeyValuePair();
                        }
                        break;
                    case states.FULL_MESSAGE_FINISHED:
                        // clear out the amp_msg_raw map so we can fill it w/ the next one
                        amp_msg_raw.Clear();
                        state = states.KEY_LEN_READ;
                        break;
                }
            }
        }

        public void processFullMessage(AMP_Msg_Raw raw_msg)
        {
            //if (fxStream == null) return; // stop here if under testing

            if ( raw_msg.ContainsKey("_command") )
            {
                string command = Encoding.UTF8.GetString(raw_msg["_command"]);
                processCommand(command, raw_msg);
            } else if ( raw_msg.ContainsKey("_answer") ) {
                processAnswer(raw_msg);
            } else if ( raw_msg.ContainsKey("_error") ) {
                processError(raw_msg);
            } else {
                log.Error("Got AMP message w/o a _command, _answer, or _error key!");
                return;
            }
        }

        public void processCommand(string cmdName, AMP_Msg_Raw raw_msg)
        {
            log.Debug("Processing AMP command: " + cmdName);

            if (registeredCommands.ContainsKey(cmdName))
            {
                // YES, THIS COMMAND HAS A REGISTERED RESPONDER.
                responderStruct respRef = registeredCommands[cmdName];
                AMP_Msg requestMsg = respRef.command.toTypedMessage(raw_msg, AMP_Command.Type.REQUEST);
                try
                {
                    Object returnVal = null;
                    if (respRef.callbackContext != null)
                    {
                        respRef.callbackContext.Send(delegate
                             {
                                 returnVal = respRef.responder(requestMsg, respRef.state);
                             }, null);
                    }
                    else if (respRef.callbackControl != null)
                    {
                        returnVal = respRef.callbackControl.Invoke(respRef.responder, requestMsg, respRef.state);
                    }
                    else
                    {
                        returnVal = respRef.responder(requestMsg, respRef.state);
                    }

                    if (!raw_msg.ContainsKey("_ask")) return;

                    // try to cast returnVal to AMP_Msg first, then if that fails cast it to
                    // DelegateBox since those are the only supported return types from an AMP responder
                    try
                    {
                        var responseMsg = (AMP_Msg)returnVal;
                        _sendAMPResponse(responseMsg, respRef.command, raw_msg["_ask"]);
                    }
                    catch (InvalidCastException)
                    {
                        byte[] ask_key = raw_msg["_ask"];
                        var dBox = (DelegateBox)returnVal;
                        dBox.setCallback(response => _processDeferredAMPResponse(response, respRef.command, ask_key));
                    }
                }
                catch (Exception e)
                {
                    if (raw_msg.ContainsKey("_ask"))
                    {
                        _sendAMPError(e, respRef.command, raw_msg["_ask"]);
                    }
                }
            }
            else
            {
                // NO, THERE ISN'T ANY REGISTERED RESPONDER FOR THIS COMMAND
                log.Error("No such AMP command: " + cmdName);
                if (raw_msg.ContainsKey("_ask"))
                {
                    var response = new AMP_Msg_Raw
                                       {
                                           {"_error", raw_msg["_ask"]},
                                           {"_error_code", AMP_Command.UNHANDLED_ERROR_CODE},
                                           {
                                               "_error_description",
                                               Encoding.UTF8.GetBytes(String.Format("Unhandled Command: {0}", cmdName))
                                               }
                                       };
                    SendMessage(response);
                }
                return;
            }
        }

        private void _sendAMPError(Exception e, AMP_Command cmd, byte[] ask_key)
        {
            var response = new AMP_Msg_Raw {{"_error", ask_key}};
            byte[] error_code;
            byte[] descr;
            if (cmd.errors.ContainsKey(e.GetType()))
            {
                error_code = cmd.errors[e.GetType()];
                descr = Encoding.UTF8.GetBytes(e.ToString());
            }
            else
            {
                error_code = AMP_Command.UNKNOWN_ERROR_CODE;
                descr = Encoding.UTF8.GetBytes("Unknown Error");
            }
            response.Add("_error_code", error_code);
            response.Add("_error_description", descr);
            SendMessage(response);
        }

        private void _sendAMPResponse(AMP_Msg msg, AMP_Command cmd, byte[] answer_key)
        {
            AMP_Msg_Raw responseMsg = cmd.toRawMessage(msg, AMP_Command.Type.RESPONSE);
            responseMsg["_answer"] = answer_key;
            SendMessage(responseMsg);
        }

        public virtual void SendMessage(AMP_Msg_Raw raw_msg)
        {
            transport.Write(buildAMPWireCommand(raw_msg));
        }

        private void _processDeferredAMPResponse(Object response, AMP_Command cmd, byte[] answer_key)
        {
            try
            {
                try
                {
                    var msg = (AMP_Msg)response;
                    _sendAMPResponse(msg, cmd, answer_key);
                }
                catch (InvalidCastException)
                {
                    // assume it's an exception
                    var err = (Exception) response;
                    _sendAMPError(err, cmd, answer_key);
                }
            }
            catch (Exception e)
            {
                _sendAMPError(e, cmd, answer_key);
            }
        }

        public void processAnswer(AMP_Msg_Raw raw_msg)
        {
            int answerKey = Int32.Parse(Encoding.UTF8.GetString(raw_msg["_answer"]));
            log.Debug("Processing AMP answer for request " + answerKey);
            if ( ! askInfoDict.ContainsKey(answerKey) ) {
                log.ErrorFormat("Got AMP answer message w/o an unknown key {0}! (Perhaps we already timed out?)", answerKey);
                return;
            }
            Ask_Info askInfo = askInfoDict[answerKey];
            askInfo.ar._result = askInfo.cmd.toTypedMessage(raw_msg, AMP_Command.Type.RESPONSE);
            try
            {
                //askInfo.ar._callback(askInfo.ar);
                askInfo.ar.doCallback();

            } catch (Exception e) {
                // XXX TODO should this catch exceptions? We aren't really writing a *framework*, a la Twisted, here. So since we're just
                // writing a *library*, maybe it should be up to the client code to deal with exceptions.
                // NOTE: We can't catch errors anyway in the case that doCallback posts the callback on the SyncronizationContext
                log.Error("Exception raised in user-defined AMP result callback:", e);
            } finally {
                askInfoDict.Remove(answerKey);
            }
        }

        public void processError(AMP_Msg_Raw raw_msg)
        {
            // According to the AMP spec an error packet should contain all 3 of these keys
            int msgKey = Int32.Parse(Encoding.UTF8.GetString(raw_msg["_error"]));

            byte[] errorCode = raw_msg["_error_code"];
            string description = Encoding.UTF8.GetString(raw_msg["_error_description"]);

            log.DebugFormat("Procesing AMP error for request {0} code: {1} descr: {2}", msgKey, Bytes.ToHexStr(errorCode), description);

            if ( ! askInfoDict.ContainsKey(msgKey) ) {
                log.ErrorFormat("Got AMP error message w/ an unknown key {0}! (Perhaps we already timed out?)", msgKey);
                return;
            }

            Ask_Info askInfo = askInfoDict[msgKey];

            Type excType = typeof(UnknownRemoteError); // default if we don't find the _error_code below

            foreach (var kvp in askInfo.cmd.errors)
            {
                if (Bytes.AreEqual(kvp.Value, errorCode))
                {
                    excType = kvp.Key;
                }
            }

            var exc = (RemoteAmpError)Activator.CreateInstance(excType, errorCode, description);

            askInfo.ar._error = exc;
            try
            {
                askInfo.ar.doCallback();
            } catch (Exception e) {
                // XXX TODO should this catch exceptions? We aren't really writing a *framework*, a la Twisted, here. So since we're just
                // writing a *library*, maybe it should be up to the client code to deal with exceptions.
                log.Error("AMP error returned, and exception raised in user-defined callback:", e);
            } finally {
                askInfoDict.Remove(msgKey);
            }
        }

        // If we ever allow deferred responses we might need to synchronize this method
        public void writeEmptyResponse( byte[] key )
        {
            var response = new AMP_Msg_Raw {{"_answer", key}};
            SendMessage(response);
        }

        public void processCurrentKeyValuePair()
        {
            var key_data_array = (byte[])key_data.ToArray(typeof(byte));
            //byte[] val_data_array = (byte[])val_data.ToArray(typeof(byte));
            //amp_msg_raw.Add(Encoding.UTF8.GetString(key_data_array), Encoding.UTF8.GetString(val_data_array));
            //amp_msg_raw.Add(Encoding.UTF8.GetString(key_data_array), Encoding.UTF8.GetString(val_data));
            amp_msg_raw.Add(Encoding.UTF8.GetString(key_data_array), val_data);

            key_len.Clear();
            key_data.Clear();

            val_len.Clear();
            //val_data.Clear();
            val_data = new byte[0];

            val_len_int = 0;
        }

        public static byte[] buildAMPWireCommand(AMP_Msg_Raw amp_msg)
        {
            //byte[] result2 = {};
            //return result2;

            var tmp_result = new ArrayList();
            byte[] value;

            if (amp_msg.Count == 0)
            {
                throw new AMPMessageEmptyError();
            }
            if (BitConverter.IsLittleEndian)
            {
                foreach (string key in amp_msg.Keys)
                {
                    value = amp_msg[key];
                    if (key.Length == 0 || key.Length > 255) throw new AMPInvalidKeySizeError();
                    if (value.Length > 65535) throw new AMPInvalidValueSizeError();
                    tmp_result.Add((byte)0x00); // first byte of key length
                    // use the first byte of the 2-byte sequence, and store that as the 2nd byte in the tmp_result
                    // this is because the host computer is little-endien, but we're encoding wire values in network byte order (big-endien)
                    tmp_result.Add(BitConverter.GetBytes((UInt16)key.Length)[0]);
                    tmp_result.AddRange(Encoding.UTF8.GetBytes(key));
                    byte[] tmp = BitConverter.GetBytes((UInt16)value.Length);
                    tmp_result.Add(tmp[1]);
                    tmp_result.Add(tmp[0]);
                    tmp_result.AddRange(value);
                }
            }
            else
            {
                foreach (string key in amp_msg.Keys)
                {
                    value = amp_msg[key];
                    if (key.Length == 0 || key.Length > 255) throw new AMPInvalidKeySizeError();
                    if (value.Length > 65535) throw new AMPInvalidValueSizeError();
                    tmp_result.Add((byte)0x00); // first byte of key length

                    tmp_result.Add(BitConverter.GetBytes((UInt16)key.Length)[1]);
                    tmp_result.AddRange(Encoding.UTF8.GetBytes(key));
                    byte[] tmp = BitConverter.GetBytes((UInt16)value.Length);
                    tmp_result.Add(tmp[0]);
                    tmp_result.Add(tmp[1]);
                    tmp_result.AddRange(value);
                }
            }
            // terminated w/ an empty key
            tmp_result.Add( (byte)0x00 );
            tmp_result.Add( (byte)0x00 );

            //byte[] result = {1};
            //return (byte[])tmp_result;
            //return tmp_result.ToArr
            var result = (byte[])tmp_result.ToArray(typeof(byte));
            return result;
        }

        public AMP_Msg CallRemote(AMP_Command cmd, AMP_Msg msg)
        {
            IAsyncResult ar = BeginCallRemote(cmd, msg, null, null);
            ar.AsyncWaitHandle.WaitOne();
            return EndCallRemote(ar);
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        public IAsyncResult BeginCallRemote(AMP_Command cmd, AMP_Msg msg, AsyncCallback callback, Object cbState)
        {
            var ar = new SimpleAsyncResult {_callback = callback, _asyncState = cbState};
            int currentKey = ++lastAskKey;
            //askInfoDict.Add(currentKey, new Ask_Info(currentKey, cmd, callback, ar));
            askInfoDict.Add(currentKey, new Ask_Info(currentKey, cmd, ar));
            AMP_Msg_Raw raw_msg = cmd.toRawMessage(msg, AMP_Command.Type.REQUEST);
            raw_msg.Add("_ask", Encoding.UTF8.GetBytes(currentKey.ToString()));
            raw_msg.Add("_command", Encoding.UTF8.GetBytes(cmd.name));
            log.DebugFormat("Calling remote AMP Command: {0} with key {1}", cmd.name, currentKey);
            SendMessage(raw_msg);

            // set up timeout in case server never responds

            // C# is fucking retarded and makes you keep explicit references to this sort of object
            int timerKey = lastTempTimerKey;
            lastTempTimerKey += 1;
            var t = new Timer(delegate { remoteCallTimeoutCheck(currentKey, timerKey); },
                              null, (int) (callRemoteTimeout*1000.0), Timeout.Infinite);
            tempTimers.Add(timerKey, t); // remember the Timer

            return ar;
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        public void remoteCallTimeoutCheck(int askKey, int timerKey)
        {
            tempTimers.Remove(timerKey); // forget about the Timer
            if (askInfoDict.ContainsKey(askKey))
            {
                Ask_Info ai = askInfoDict[askKey];
                askInfoDict.Remove(askKey);
                log.Error("Timed out waiting for AMP response for request " + askKey);
                ai.ar._error = new ServerTimeout();
                try
                {
                    ai.ar.doCallback();
                }
                catch (Exception e)
                {
                    log.Error("AMP call timed out, and exception raised in user-defined callback:", e);
                }
            }
        }

        public AMP_Msg EndCallRemote(IAsyncResult ar)
        {
            SimpleAsyncResult result = throwIfError(ar);

            // No error, we must have a result from the server available, then
            return (AMP_Msg)result._result;
        }

        public static SimpleAsyncResult throwIfError(IAsyncResult ar)
        {
            var result = (SimpleAsyncResult)ar;
            if (result._error != null)
            {
                throw result._error;
            }
            return result;
        }
    }

    public class AMP_Type
    {
        public string name;
        public AMP_Type(string name)
        {
            this.name = name;
        }
        //We're cheating here and actually using strings to store the serialized result
        //It's OK because we're never going to be handing non-ASCII data and so we'll just
        //use the UTF8 encoder to get the actual bytes later on
        public virtual byte[] toBytes(Object obj)
        {
            return new byte[0];
        }
        public virtual Object fromBytes(byte[] bytes)
        {
            return new Object();
        }
    }

    public class AMP_String : AMP_Type
    {
        public AMP_String()
        : base("string")
        {
        }
        public override byte[] toBytes(Object obj)
        {
            return Encoding.UTF8.GetBytes((string)obj);
        }
        public override Object fromBytes(byte[] bytes)
        {
            return Encoding.UTF8.GetString(bytes);
        }
    }

    public class AMP_Bytes : AMP_Type
    {
        public AMP_Bytes()
            : base("bytes")
        {
        }
        public override byte[] toBytes(Object obj)
        {
            return (byte[])obj;
        }
        public override Object fromBytes(byte[] bytes)
        {
            return bytes;
        }
    }

    public class AMP_Bool : AMP_Type
    {
        public byte[] TRUE =  Encoding.UTF8.GetBytes("True");
        public byte[] FALSE = Encoding.UTF8.GetBytes("False");

        public AMP_Bool()
        : base("bool")
        {
        }
        public override byte[] toBytes(Object obj)
        {
			return (bool)obj ? TRUE : FALSE;
        }
        public override Object fromBytes(byte[] bytes)
        {
            if (Bytes.AreEqual(TRUE, bytes))
            {
                return true;
            }
            if (Bytes.AreEqual(FALSE, bytes))
            {
                return false;
            }
            /// XXX If this throws it blows up when passing this error across thread boundaries
            /// with SerializationException.  test_AMP_error will demonstrate if you
            /// change this method to always throw
            throw new AMPTypeDecodeError();
        }
    }

    public class AMP_Integer : AMP_Type
    {
        public AMP_Integer()
        : base("int")
        {
        }
        public override byte[] toBytes(Object obj)
        {
            return Encoding.UTF8.GetBytes(((int)obj).ToString());
        }
        public override Object fromBytes(byte[] bytes)
        {
            return Int32.Parse(Encoding.UTF8.GetString(bytes));
        }
    }

    public class AMP_Decimal : AMP_Type
    {
        public AMP_Decimal()
        : base("decimal")
        {
        }
        public override byte[] toBytes(Object obj)
        {
            return Encoding.UTF8.GetBytes(((decimal)obj).ToString());
        }
        public override Object fromBytes(byte[] bytes)
        {
            return decimal.Parse(Encoding.UTF8.GetString(bytes));
        }
    }

    public class AMP_TimeArgument : AMP_Type
    {
        public AMP_TimeArgument()
        : base("TimeArgument")
        {
        }
        public override byte[] toBytes(Object obj)
        {
            DateTime dt = ((DateTime)obj).ToUniversalTime();
            string s = String.Format("{0}/{1}/{2} {3}:{4}:{5}", dt.Year, dt.Month, dt.Day, dt.Hour, dt.Minute, dt.Second);
            return Encoding.UTF8.GetBytes(s);
        }
        public override Object fromBytes(byte[] bytes)
        {
            string s = Encoding.UTF8.GetString(bytes);
            string[] parts = s.Split(' ');
            string[] date_parts = parts[0].Split('/');
            string[] time_parts = parts[1].Split(':');

            int year = int.Parse(date_parts[0]);
            int month = int.Parse(date_parts[1]);
            int day = int.Parse(date_parts[2]);

            int hour = int.Parse(time_parts[0]);
            int min = int.Parse(time_parts[1]);
            int sec = int.Parse(time_parts[2]);

            var dt = new DateTime(year, month, day, hour, min, sec, DateTimeKind.Utc);
            return dt;
        }
    }

    public class AMP_Parameter
    {
        public string name;
        public AMP_Type type;

        public AMP_Parameter(string name, AMP_Type ampType)
        {
            this.name = name;
            type = ampType;
        }
    }

    public class AMP_Command
    {
		private static readonly log4net.ILog log = log4net.LogManager.GetLogger("AMP");
        public enum Type
        {
            REQUEST,
            RESPONSE
        }

        public bool requiresAnswer = true;

        public string name;
        public List<AMP_Parameter> arguments;
        public List<AMP_Parameter> response;
        public Dictionary<System.Type,byte[]> errors = new Dictionary<System.Type,byte[]>();

        public static byte[] UNKNOWN_ERROR_CODE = Encoding.UTF8.GetBytes("UNKNOWN");
        public static byte[] UNHANDLED_ERROR_CODE = Encoding.UTF8.GetBytes("UNHANDLED");

        // XXX TODO make sure FXStream subclasses add this since we are removing it from here
        //public static byte[] UNAUTHORIZED_ERROR_CODE = Encoding.UTF8.GetBytes("UNAUTHORIZED");

        public AMP_Command(string name)
        {
            this.name = name;
            arguments = new List<AMP_Parameter>();
            response = new List<AMP_Parameter>();

            // These are the errors that every AMP_Command will know about, since they
            // are part of the AMP protocol.
            errors.Add(typeof(UnknownRemoteError), UNKNOWN_ERROR_CODE);
            errors.Add(typeof(UnhandledCommand), UNHANDLED_ERROR_CODE);

            // Unathorized is a RemoteAmpError that I'm going to make all my AMP_Commands support
            //errors.Add(typeof(Unauthorized), UNAUTHORIZED_ERROR_CODE);
        }

        public AMP_Msg_Raw toRawMessage(AMP_Msg amp_typed_msg, Type msgType)
        {
            List<AMP_Parameter> paramList = (msgType == Type.REQUEST ? arguments : response);
            var result = new AMP_Msg_Raw();
            foreach(AMP_Parameter param in paramList)
            {
                if (!amp_typed_msg.ContainsKey(param.name))
                {
                    log.ErrorFormat("Command {0}: paramter \"{1}\" not found in msg!", name, param.name);
                    throw new AMPParameterNotFound();
                }
                result.Add( param.name, param.type.toBytes(amp_typed_msg[param.name]));
            }
            return result;
            }

        public AMP_Msg toTypedMessage(AMP_Msg_Raw amp_raw_msg, Type msgType)
        {
            List<AMP_Parameter> paramList = (msgType == Type.REQUEST ? arguments : response);
            var result = new AMP_Msg();
            foreach(AMP_Parameter param in paramList)
            {
                if (!amp_raw_msg.ContainsKey(param.name))
                {
                    log.ErrorFormat("Command {0}: paramter \"{1}\" not found in msg!", name, param.name);
                    throw new AMPParameterNotFound();
                }
                result.Add( param.name, param.type.fromBytes(amp_raw_msg[param.name]));
            }
            return result;
            }
        }

    public class Ask_Info
    {
        public int askKey;
        public AMP_Command cmd;
        //public AsyncCallback callback;
        public SimpleAsyncResult ar;

        //public Ask_Info(int askKey, AMP_Command cmd, AsyncCallback callback, FXStreamAsyncResult ar)
        public Ask_Info(int askKey, AMP_Command cmd, SimpleAsyncResult ar)
        {
            this.askKey = askKey;
            this.cmd = cmd;
            //this.callback = callback;
            this.ar = ar;
        }

    }

    // Container class for a raw AMP "box"... this always maps field names to byte arrays.
    // We convert back and forth between this class and the TYPED version of an AMP box, AMP_Msg,
    // using the Command class which knows how to map fields to types.
    public class AMP_Msg_Raw : Dictionary<string, byte[]>
    {
        internal static void printBytes(byte[] bytes)
        {
            foreach (byte b in bytes)
            {
                Console.Write(String.Format("\\x{0:x2}", b));
            }
            Console.WriteLine();
        }

        public override bool Equals(object obj)
        {
            if (!(obj is AMP_Msg_Raw))
            {
                return false;
            }
            return this == (AMP_Msg_Raw)obj;
           
        }

        public static bool operator ==(AMP_Msg_Raw a, AMP_Msg_Raw b)
        {
            if (a.Count != b.Count)
            {
                return false;
            }
            foreach (string key in a.Keys) {
                if ( ! b.ContainsKey(key)) {
                    return false;
                }
                if (a[key].Length != b[key].Length) return false;
                for (int i = 0; i < a[key].Length; i++)
                {
                    if (a[key][i] != b[key][i]) return false;
                }
            }
            return true;
        }


        public static bool operator !=(AMP_Msg_Raw a, AMP_Msg_Raw b)
        {
            return !(a == b);
        }

        // TODO is this supposed to test value equality or object identity?
        // This implementation is the default that ReSharper gave me
        public bool Equals(AMP_Msg_Raw msg)
        {
            return !ReferenceEquals(null, msg);
        }

        public override int GetHashCode()
        {
            return 0;
        }
    }

    // Container class for a TYPED AMP message. The actual values will be instances of the specific
    // types that have been defined for the fields. This class is what client code will use to handle
    // AMP requests and responses.
    public class AMP_Msg : Dictionary<string, Object>
    {
    }

    public class SimpleAsyncResult : IAsyncResult
    {
        // TODO these really should be defined 'internal'
        public object _result;
        public Exception _error;
        public AsyncCallback _callback;
        public SynchronizationContext _savedContext;

        public SimpleAsyncResult()
        {
            _savedContext = SynchronizationContext.Current;
            _waitHandle = new ManualResetEvent(false);
        }

        public void ThrowIfError()
        {
            if (_error != null)
            {
                throw _error;
            }
        }

        // TODO this really should be defined 'internal'
        public void doCallback()
        {
            _isCompleted = true;
            try
            {
                if (_callback != null)
                {
                    if (_savedContext != null)
                    {
                        // Post the callback to the context
                        // (Callback will happen asynchronously in the saved context's thread)
                        // XXX TODO shouldn't we pass _asyncState to the delegate here, and also
                        // pass it to the callback in the case where we run the callback synchronously (below) ?
                        // maybe not, because AsyncCallback is only defined to take a single argument: IAsyncResult
                        _savedContext.Post(delegate { _callback(this); }, null);
                    }
                    else
                    {
                        // Just callback synchronously
                        _callback(this);
                    }
                }
            }
            finally
            {
                _waitHandle.Set();
            }
        }

        // BEGIN IAsyncResult properties

        // TODO these really should be defined 'internal'
        public bool _isCompleted;
        public Object _asyncState;
        public bool _completedSynchronously;
        public ManualResetEvent _waitHandle;

        public Object AsyncState
        {
            get
            {
                return _asyncState;
            }
        }

        public WaitHandle AsyncWaitHandle
        {
            get
            {
                return _waitHandle;
            }
        }

        public bool IsCompleted {
            get
            {
                return _isCompleted;
            }
        }
        public bool CompletedSynchronously {
            get
            {
                return _completedSynchronously;
            }
        }

        // END IAsyncResult properties

    }

    public class SocketClosedException : SocketException
    {
    }

    public class AmpError : Exception
    {
        public AmpError(string description)
            : base(description)
        {
        }

        public AmpError()
        {
        }
    }

    public class AMPProtocolError : AmpError
    {
    }
    public class AMPMessageEmptyError : AmpError
    {
    }
    public class AMPInvalidKeySizeError : AmpError
    {
    }
    public class AMPInvalidValueSizeError : AmpError
    {
    }
    public class AMPParameterNotFound : AmpError
    {
    }
    public class AMPTypeDecodeError : AmpError
    {
    }
    
    public class RemoteAmpError : AmpError
    {
        public byte[] errorCode;
        public RemoteAmpError() { }
        public RemoteAmpError(byte[] errorCode, string description)
            : base(description)
        {
            this.errorCode = errorCode;
        }
    }

    internal class Bytes
    {
        public static bool AreEqual(byte[] x, byte[] y)
        {
            if (x.Length != y.Length) return false;
            for (int i = 0; i < x.Length; i++)
            {
                if (x[i] != y[i]) return false;
            }
            return true;
        }
        public static string ToHexStr(byte[] bytes)
        {
            var sb = new StringBuilder();
            foreach (byte b in bytes)
            {
                if (b < 32 || b > 126)
                {
                    sb.AppendFormat("\\x{0:x2}", b);
                }
                else
                {
                    if (b == '\\')
                    {
                        sb.Append("\\\\");
                    }
                    else
                    {
                        sb.Append((char)b);
                    }
                }
            }
            return sb.ToString();
        }
    }

    public class UnknownRemoteError: RemoteAmpError
    {
        public UnknownRemoteError(byte[] errorCode, string description)
            : base(errorCode, description)
        {
        }
    }

    public class UnhandledCommand: RemoteAmpError
    {
        public UnhandledCommand(byte[] errorCode, string description)
            : base(errorCode, description)
        {
        }
    }

    public class ServerTimeout : AmpError
    {
    }
}
