﻿using System.Net;

namespace AMPTests
{
    using System;
    using NUnit.Framework;
#if NUNIT_24
    using NUnit.Framework.SyntaxHelpers;
#endif
    using System.Collections;
    using System.Collections.Generic;
    using System.Text;
    using System.Windows.Forms;
    using System.Threading;
    using System.Net.Sockets;

    using AMP;

    public delegate byte[] stringToBytes(string foo);

    public class MySillyForm : Form
    {
        public int formThreadId;
        public int callbackThreadId;

        public ManualResetEvent waitHandle;
        public IAsyncResult ar;

        public MySillyForm(ManualResetEvent waitHandle)
        {
            this.waitHandle = waitHandle;
            Load += MySillyForm_Load;
        }

        public void MySillyForm_Load(object sender, EventArgs e)
        {
            formThreadId = Thread.CurrentThread.ManagedThreadId;

            Console.WriteLine("Form thread {0}", formThreadId);
            Console.WriteLine(String.Format("Form ctx: {0}", SynchronizationContext.Current));

            // when we supply any delegates to an AMP method the current
            // synchronization context will be captured. This means that
            // calls to run those delegates will be posted to the synchronization
            // context's message loop, and will thus execute in the same thread
            // that the original AMP call was made in.

            var result = new SimpleAsyncResult();
            result._callback = delegate
            {
                callbackThreadId = Thread.CurrentThread.ManagedThreadId;
                waitHandle.Set();
            };

            // run the callback on a foreign thread, as it would be if this callback were
            // the result of receiving network traffic
            var t = new Thread(result.doCallback);
            t.Start();
        }
    }

    [TestFixture]
    public class StringDictsTests
    {
        public stringToBytes _b = foo => Encoding.UTF8.GetBytes(foo);

        internal AMP_Msg_Raw a;
        internal AMP_Msg_Raw b;
        [SetUp]
        public void Init()
        {
            a = new AMP_Msg_Raw();
            b = new AMP_Msg_Raw();
        }

        [Test]
        public void empty()
        {
            Assert.That(a, Is.EqualTo(b));
        }

        [Test]
        public void diff_size()
        {
            // a stays empty
            b.Add("key", _b("value"));
            Assert.That(a, Is.Not.EqualTo(b));
        }

        [Test]
        public void diff_keys()
        {
            a.Add("key", _b("value"));
            b.Add("key2", _b("value"));
            Assert.That(a, Is.Not.EqualTo(b));
        }

        [Test]
        public void diff_values()
        {
            a.Add("key", _b("value"));
            b.Add("key", _b("value2"));
            Assert.That(a, Is.Not.EqualTo(b));
        }

        [Test]
        public void equal_size_1()
        {
            a.Add("key", _b("value"));
            b.Add("key", _b("value"));
            Assert.That(a, Is.EqualTo(b));
        }

        [Test]
        public void equal_size_2()
        {
            a.Add("key", _b("value"));
            a.Add("key2", _b("value2"));
            b.Add("key", _b("value"));
            b.Add("key2", _b("value2"));
            Assert.That(a, Is.EqualTo(b));
        }
    }

    [TestFixture]
    public class BuildAMPWireCommand_Tests
    {
        public stringToBytes _b = foo => Encoding.UTF8.GetBytes(foo);
        //[SetUp]
        //public void Init()
        //{
        //}
        [Test]
        [ExpectedException(typeof(AMPMessageEmptyError))]
        public void build_msg_empty()
        {
            AMP_Protocol.buildAMPWireCommand(new AMP_Msg_Raw());
        }

        [Test]
        [ExpectedException(typeof(AMPInvalidKeySizeError))]
        public void build_msg_null_key()
        {
            var amp_msg = new AMP_Msg_Raw { { "", _b("ignored") } };
            AMP_Protocol.buildAMPWireCommand(amp_msg);
        }

        [Test]
        [ExpectedException(typeof(AMPInvalidKeySizeError))]
        public void build_msg_key_too_long()
        {
            var amp_msg = new AMP_Msg_Raw();
            const string chunk = "0123456789ABCDEF";
            string key = "";
            for (int i = 0; i < 16; i++)
            { // 16 bytes 16 times is 256 bytes which is 1 byte too long
                key += chunk;
            }
            Assert.That(key.Length, Is.EqualTo(256));
            amp_msg.Add(key, _b("ignored"));
            AMP_Protocol.buildAMPWireCommand(amp_msg);
        }

        [Test]
        public void build_msg_max_key_len()
        // this also doubles as a test for checking that empty
        // values in the StringDict are handled OK
        {
            var amp_msg = new AMP_Msg_Raw();
            const string chunk = "0123456789ABCDEF";
            string key = "";
            for (int i = 0; i < 15; i++)
            { // 16 bytes 15 times is 240 bytes
                key += chunk;
            }
            key += "0123456789ABCDE"; // 15 more bytes gives us 255
            Assert.That(key.Length, Is.EqualTo(255));

            amp_msg.Add(key, _b(""));
            byte[] result = AMP_Protocol.buildAMPWireCommand(amp_msg);
            var expected_list = new ArrayList { (byte)0x00, (byte)0xff };
            expected_list.AddRange(Encoding.UTF8.GetBytes(key));
            expected_list.Add((byte)0x00);
            expected_list.Add((byte)0x00);
            expected_list.Add((byte)0x00);
            expected_list.Add((byte)0x00);
            var expected = (byte[])expected_list.ToArray(typeof(byte));
            Assert.That(expected, Is.EqualTo(result));
        }

        [Test]
        [ExpectedException(typeof(AMPInvalidValueSizeError))]
        public void build_msg_value_too_long()
        {
            var amp_msg = new AMP_Msg_Raw();
            const string chunk = "0123456789ABCDEF";
            string value = "";
            for (int i = 0; i < 4096; i++)
            { // 16 bytes 4096 times is 65536 bytes (1 byte too long)
                value += chunk;
            }
            Assert.That(value.Length, Is.EqualTo(65536));
            amp_msg.Add("key", _b(value));
            AMP_Protocol.buildAMPWireCommand(amp_msg);
        }

        [Test]
        public void build_msg_max_value_len()
        {
            var amp_msg = new AMP_Msg_Raw();
            const string chunk = "0123456789ABCDEF";
            string value = "";
            for (int i = 0; i < 4095; i++)
            { // 16 bytes 15 times is 65520 bytes
                value += chunk;
            }
            value += "0123456789ABCDE"; // 15 more bytes gives us 65535
            Assert.That(value.Length, Is.EqualTo(65535));

            amp_msg.Add("k", _b(value));
            byte[] result = AMP_Protocol.buildAMPWireCommand(amp_msg);
            var expected_list = new ArrayList { (byte)0x00, (byte)0x01, (byte)0x6B, (byte)0xff, (byte)0xff };
            expected_list.AddRange(Encoding.UTF8.GetBytes(value));
            expected_list.Add((byte)0x00);
            expected_list.Add((byte)0x00);
            var expected = (byte[])expected_list.ToArray(typeof(byte));
            Assert.That(expected, Is.EqualTo(result));
        }
    }

    public class Test_AMP_Protocol : AMP_Protocol
    {
        public List<AMP_Msg_Raw> messages = new List<AMP_Msg_Raw>();
        public override void SendMessage(AMP_Msg_Raw raw_msg)
        {
            messages.Add(raw_msg);
        }
    }

    /// <summary>
    /// Test the actual AMP wire-protcol parsing and dispatching.
    /// XXX TODO need to test that processCommand() processAnswer() and processError() is called
    /// correctly... we probably want to override these method and simply capture the arguments
    /// they are called with.
    /// </summary>
    [TestFixture]
    public class AMP_Parser_Tests
    {
        public stringToBytes _b = foo => Encoding.UTF8.GetBytes(foo);

        internal AMP_Protocol proto;
        [SetUp]
        public void Init()
        {
            proto = new Test_AMP_Protocol();
        }

        [Test]
        [ExpectedException(typeof(AMPProtocolError))]
        public void parse_initial_null()
        {
            byte[] chunk = { 0, 0 };
            proto.DataReceived(chunk);
        }

        [Test]
        [ExpectedException(typeof(AMPProtocolError))]
        public void parse_key_too_big()
        {
            byte[] chunk = { 1, 0 };
            proto.DataReceived(chunk);
        }

        [Test]
        public void parse_one_byte_key_value()
        {
            var amp_msg = new AMP_Msg_Raw { { "x", _b("y") } };
            byte[] bytes = AMP_Protocol.buildAMPWireCommand(amp_msg);
            proto.DataReceived(bytes);
            Assert.That(amp_msg, Is.EqualTo(proto.amp_msg_raw));
        }

        [Test]
        public void parse_one_byte_key_empty_value()
        {
            var amp_msg = new AMP_Msg_Raw { { "x", _b("") } };
            byte[] bytes = AMP_Protocol.buildAMPWireCommand(amp_msg);
            proto.DataReceived(bytes);
            foreach (byte b in bytes)
            {
                Console.Write(String.Format("\\x{0:x2}", b));
            }
            Console.WriteLine();

            // NUnit kinda sucks a lot. This doesn't work at all for no apparent reason.
            Assert.That(amp_msg, Is.EqualTo(proto.amp_msg_raw));
        }

        [Test]
        public void parse_simple_msg()
        {
            var amp_msg = new AMP_Msg_Raw
                              {
                                  {"_ask", _b("14253")},
                                  {"_command", _b("Sum")},
                                  {"x", _b("15")},
                                  {"y", _b("45")}
                              };
            byte[] bytes = AMP_Protocol.buildAMPWireCommand(amp_msg);
            proto.DataReceived(bytes);
            Assert.That(amp_msg, Is.EqualTo(proto.amp_msg_raw));
        }

        [Test]
        public void parse_simple_msg_byte_at_a_time()
        {
            var amp_msg = new AMP_Msg_Raw
                              {
                                  {"_ask", _b("14253")},
                                  {"_command", _b("Sum")},
                                  {"x", _b("15")},
                                  {"y", _b("45")}
                              };
            byte[] bytes = AMP_Protocol.buildAMPWireCommand(amp_msg);
            var buf = new byte[1];
            foreach (byte b in bytes)
            {
                buf[0] = b;
                proto.DataReceived(buf);
            }
            Assert.That(amp_msg, Is.EqualTo(proto.amp_msg_raw));
        }

        [Test]
        public void parse_double_msg()
        {
            var amp_msg = new AMP_Msg_Raw
                              {
                                  {"_ask", _b("14253")},
                                  {"_command", _b("Sum")},
                                  {"x", _b("15")},
                                  {"y", _b("45")}
                              };

            var amp_msg2 = new AMP_Msg_Raw
                               {
                                   {"_ask", _b("321")},
                                   {"_command", _b("Power")},
                                   {"x", _b("51")},
                                   {"y", _b("54")}
                               };

            byte[] bytes = AMP_Protocol.buildAMPWireCommand(amp_msg);
            byte[] bytes2 = AMP_Protocol.buildAMPWireCommand(amp_msg2);

            proto.DataReceived(bytes);
            Assert.That(amp_msg, Is.EqualTo(proto.amp_msg_raw));
            proto.DataReceived(bytes2);
            Assert.That(amp_msg2, Is.EqualTo(proto.amp_msg_raw));

            var bytes3 = new byte[bytes.Length + bytes2.Length];
            bytes2.CopyTo(bytes3, 0);
            bytes.CopyTo(bytes3, bytes2.Length);

            proto.DataReceived(bytes3);
            Assert.That(amp_msg, Is.EqualTo(proto.amp_msg_raw));
        }
    }

    internal class AMP_TEST_Command : AMP_Command
    {
        public AMP_TEST_Command()
            : base("test")
        {
            arguments.Add(new AMP_Parameter("one", new AMP_String()));
            arguments.Add(new AMP_Parameter("two", new AMP_String()));

            response.Add(new AMP_Parameter("result", new AMP_String()));
        }
    }

    [TestFixture]
    public class AMP_Command_Tests
    {
        public stringToBytes _b = foo => Encoding.UTF8.GetBytes(foo);

        internal AMP_Command cmd;
        internal AMP_Command test_cmd;
        [SetUp]
        public void Init()
        {
            cmd = new AMP_Command("cmd");
            test_cmd = new AMP_TEST_Command();
        }
        [Test]
        public void toRawMessage_request_empty()
        {
            var amp_typed_msg = new AMP_Msg();
            AMP_Msg_Raw result = cmd.toRawMessage(amp_typed_msg, AMP_Command.Type.REQUEST);
            Assert.That(result.Count, Is.EqualTo(0));
        }
        [Test]
        public void toRawMessage_response_empty()
        {
            var amp_typed_msg = new AMP_Msg();
            AMP_Msg_Raw result = cmd.toRawMessage(amp_typed_msg, AMP_Command.Type.RESPONSE);
            Assert.That(result.Count, Is.EqualTo(0));
        }

        [Test]
        [ExpectedException(typeof(AMPParameterNotFound))]
        public void toRawMessage_request_missing_parameter()
        {
            var amp_typed_msg = new AMP_Msg();
            test_cmd.toRawMessage(amp_typed_msg, AMP_Command.Type.REQUEST);
        }
        [Test]
        [ExpectedException(typeof(AMPParameterNotFound))]
        public void toRawMessage_response_missing_parameter()
        {
            var amp_typed_msg = new AMP_Msg();
            test_cmd.toRawMessage(amp_typed_msg, AMP_Command.Type.RESPONSE);
        }
        [Test]
        public void toRawMessage_normal()
        {
            var amp_typed_msg = new AMP_Msg { { "one", (object)"one_value" }, { "two", (object)"two_value" } };

            AMP_Msg_Raw result = test_cmd.toRawMessage(amp_typed_msg, AMP_Command.Type.REQUEST);
            Assert.That(result.Count, Is.EqualTo(2));
            Assert.That(_b("one_value"), Is.EqualTo(result["one"]));
            Assert.That(_b("two_value"), Is.EqualTo(result["two"]));

            amp_typed_msg = new AMP_Msg { { "result", (object)"one_valuetwo_value" } };

            result = test_cmd.toRawMessage(amp_typed_msg, AMP_Command.Type.RESPONSE);
            Assert.That(result.Count, Is.EqualTo(1));
            Assert.That(_b("one_valuetwo_value"), Is.EqualTo(result["result"]));
        }

        [Test]
        public void toTypedMessage_request_empty()
        {
            var amp_raw_msg = new AMP_Msg_Raw();
            AMP_Msg result = cmd.toTypedMessage(amp_raw_msg, AMP_Command.Type.REQUEST);
            Assert.That(result.Count, Is.EqualTo(0));
        }
        [Test]
        public void toTypedMessage_response_empty()
        {
            var amp_raw_msg = new AMP_Msg_Raw();
            AMP_Msg result = cmd.toTypedMessage(amp_raw_msg, AMP_Command.Type.RESPONSE);
            Assert.That(result.Count, Is.EqualTo(0));
        }

        [Test]
        [ExpectedException(typeof(AMPParameterNotFound))]
        public void toTypedMessage_request_missing_parameter()
        {
            var amp_raw_msg = new AMP_Msg_Raw();
            test_cmd.toTypedMessage(amp_raw_msg, AMP_Command.Type.REQUEST);
        }
        [Test]
        [ExpectedException(typeof(AMPParameterNotFound))]
        public void toTypedMessage_response_missing_parameter()
        {
            var amp_raw_msg = new AMP_Msg_Raw();
            test_cmd.toTypedMessage(amp_raw_msg, AMP_Command.Type.RESPONSE);
        }
        [Test]
        public void toTypedMessage_normal()
        {
            var amp_raw_msg = new AMP_Msg_Raw { { "one", _b("one_value") }, { "two", _b("two_value") } };

            AMP_Msg result = test_cmd.toTypedMessage(amp_raw_msg, AMP_Command.Type.REQUEST);
            Assert.That(result.Count, Is.EqualTo(2));
            Assert.That("one_value", Is.EqualTo(result["one"]));
            Assert.That("two_value", Is.EqualTo(result["two"]));

            amp_raw_msg = new AMP_Msg_Raw { { "result", _b("one_valuetwo_value") } };

            result = test_cmd.toTypedMessage(amp_raw_msg, AMP_Command.Type.RESPONSE);
            Assert.That(result.Count, Is.EqualTo(1));
            Assert.That("one_valuetwo_value", Is.EqualTo(result["result"]));
        }

    }

    [TestFixture]
    public class AMP_TimeArgument_Tests
    {
        public stringToBytes _b = foo => Encoding.UTF8.GetBytes(foo);

        [SetUp]
        public void Init()
        {
        }

        [Test]
        public void test_TimeArgument()
        {
            var atr = new AMP_TimeArgument();
            var dt = new DateTime(2008, 04, 16, 12, 30, 15, 200, DateTimeKind.Utc);
            byte[] value = atr.toBytes(dt);

            Console.WriteLine(Encoding.UTF8.GetString(value));

            Assert.That(_b("2008/4/16 12:30:15"), Is.EqualTo(value));

        }
    }

    [TestFixture]
    public class AMP_Callback_Tests
    {
        public MySillyForm theForm;

        [SetUp]
        public void Init()
        {
        }

        // Test that the SimpleAsyncResult class captures the current synchronization context
        // when it is created, and dispatches to the context's thread at callback-time
        [Test]
        public void test_form_synchronize()
        {
            int testThreadId = Thread.CurrentThread.ManagedThreadId;
            int callbackThreadId = 0; // assigned in callback

            var result = new SimpleAsyncResult();
            result._callback = delegate { callbackThreadId = Thread.CurrentThread.ManagedThreadId; };

            // run the callback on a foreign thread, as it would be if this callback were
            // the result of receiving network traffic
            var tmpWaitHandle = new ManualResetEvent(false);
            var t = new Thread(() => { result.doCallback(); tmpWaitHandle.Set(); });
            t.Start();

            tmpWaitHandle.WaitOne();

            Console.WriteLine("Test thread     : {0}", testThreadId);
            Console.WriteLine("Callback thread : {0}", callbackThreadId);

            // Assert our callback actually ran and saved its thread ID
            Assert.That(callbackThreadId, Is.Not.EqualTo(0));

            // Assert the callback did happen on a foreign thread
            Assert.That(callbackThreadId, Is.Not.EqualTo(testThreadId));

            // NOW, we're going to run a Form on another thread, and have that Form's code create the
            // SimpleAsyncResult. We should see that the login callback happens ON THE FORM'S THREAD, which
            // is quite different than the semantics we see above.

            //System.Console.WriteLine("Test thread {0}", Thread.CurrentThread.ManagedThreadId);
            Assert.That(SynchronizationContext.Current, Is.Null);
            var waitHandle = new ManualResetEvent(false);
            var t2 = new Thread(() => _doForm(waitHandle));
            t2.Start();
            // The form will Set() the waitHandle once it is done.
            waitHandle.WaitOne();

            Assert.That(theForm.formThreadId, Is.EqualTo(theForm.callbackThreadId));

            theForm.Invoke(new MethodInvoker(theForm.Close));
        }

        public void _doForm(ManualResetEvent waitHandle)
        {
            theForm = new MySillyForm(waitHandle);
            theForm.ShowDialog();
        }
    }

    /*
    public class DummyTransport: IAMPTransport
    {
        public void RegisterProtocol(IAMPProtocol proto)
        {
        }
        public void Start()
        {
        }
        public void Write(byte[] bytes)
        {
        }
    }

    [TestFixture]
    public class AMP_Responder_Form_Tests
    {
        public ResponderTestForm theForm;

        public static byte[] _b(string foo)
        {
            return Encoding.UTF8.GetBytes(foo);
        }

        /// <summary>
        /// Test that registering a responder captures the current
        /// synchronization context and that subsequent calls to the Responder
        /// are posted to context's message loop
        /// </summary>
        [Test]
        public void test_responder_form_synchronize()
        {
            var proto = new AMP_Protocol(new DummyTransport());
            theForm = new ResponderTestForm(proto);
            var t = new Thread(() => theForm.ShowDialog());
            t.Start();

            // wait till the Form has loaded and registered the responder
            theForm.onLoadHandle.WaitOne();

            // give the protocol some commands to process, so that it
            // calls the responders that the Form has registered
            var raw_msg = new AMP_Msg_Raw
                              {
                                  {"_command", _b("SimpleCmd")},
                                  {"_ask", _b("1")},
                                  {"test_val", _b("test")},
                              };
            proto.DataReceived(AMP_Protocol.buildAMPWireCommand(raw_msg));
            raw_msg["_command"] = _b("TestCmd");
            proto.DataReceived(AMP_Protocol.buildAMPWireCommand(raw_msg));

            // The form will Set() these wait handles when the responders runs.
            theForm.formRespondedHandle.WaitOne();
            theForm.syncCtxRespondedHandle.WaitOne();

            // verify that the responders ran on the Form's thread, not this one
            Assert.That(theForm.formThreadId, Is.EqualTo(theForm.formResponderThreadId));
            Assert.That(theForm.formThreadId, Is.EqualTo(theForm.syncCtxResponderThreadId));
            // close the Form
            theForm.Invoke(new MethodInvoker(theForm.Close));
        }

        public void _doForm(ManualResetEvent waitHandle)
        {
        }
    }

    public class ResponderTestForm : Form
    {
        public int formThreadId;
        public int formResponderThreadId;
        public int syncCtxResponderThreadId;

        public ManualResetEvent onLoadHandle = new ManualResetEvent(false);
        public ManualResetEvent formRespondedHandle = new ManualResetEvent(false);
        public ManualResetEvent syncCtxRespondedHandle = new ManualResetEvent(false);
        private AMP_Protocol proto;

        public ResponderTestForm(AMP_Protocol proto)
        {
            this.proto = proto;
            Load += ResponderTestForm_Load;
        }

        public void ResponderTestForm_Load(object sender, EventArgs e)
        {
            formThreadId = Thread.CurrentThread.ManagedThreadId;

            Console.WriteLine("Form thread {0}", formThreadId);
            Console.WriteLine(String.Format("Form ctx: {0}", SynchronizationContext.Current));

            // passing 'this' causes calls to the responder use this.Invoke(...)
            proto.RegisterResponder(new SimpleCmd(), SimpleCmdResponder, this);

            // passing the current Synchronization Contextx causes calls to the responder
            // ctx.Send(...)
            proto.RegisterResponder(new TestCmd(), SimpleCmdResponder2, SynchronizationContext.Current);

            onLoadHandle.Set();
        }

        private object SimpleCmdResponder(AMP_Msg msg, object state)
        {
            formResponderThreadId = Thread.CurrentThread.ManagedThreadId;
            formRespondedHandle.Set();
            return new AMP_Msg {{"ret_val", "hello"}};
        }

        private object SimpleCmdResponder2(AMP_Msg msg, object state)
        {
            syncCtxResponderThreadId = Thread.CurrentThread.ManagedThreadId;
            syncCtxRespondedHandle.Set();
            return new AMP_Msg {{"ret_val", "hello"}};
        }
    }
     */

    public class TestCmd : AMP_Command
    {
        public TestCmd()
            : base("TestCmd")
        {
            arguments.Add(new AMP_Parameter("test_val", new AMP_String()));
        }
    }

    /// <summary>
    /// Test that processFullMessage() does the Right Thing under all possible conditions.
    /// processFullMessage() is the processing function that handles every AMP "box" that 
    /// comes in off the wire. The "boxes" (represented by the AMP_Msg_Raw class) can 
    /// represent 3 major catagories of things:
    ///     - An AMP Command that the peer wishes to run.
    ///     - A response to an AMP Command we called on the peer.
    ///     - An error response to an AMP Command we called on the peer.
    /// </summary>
    public class AMP_Message_Dispatch_Base
    {
        // get a byte[] from a string (using UTF-8)
        public static stringToBytes _b = foo => Encoding.UTF8.GetBytes(foo);

        // Set up a command that is used by several test cases
        public static string test_cmdName = "TestCmd";
        public static string test_cmdNameNoAnswer = "TestCmdNoAnswer";
        public class MyAMPError : RemoteAmpError
        {
            public MyAMPError() { }
            public MyAMPError(byte[] errorCode, string description)
                : base(errorCode, description)
            {
            }
        }
        public class TestCmd : AMP_Command
        {
            public TestCmd()
                : base(test_cmdName)
            {
                arguments.Add(new AMP_Parameter("test_val", new AMP_String()));
                response.Add(new AMP_Parameter("ret_val", new AMP_String()));
                errors.Add(typeof(MyAMPError), _b("MY_ERROR"));
            }
        }
        /// <summary>
        /// TODO: write some tests that use this class. requiresAnswer is only a *hint*
        /// for the client-side portion of the protocol...
        /// </summary>
        public class TestCmdNoAnswer : AMP_Command
        {
            public TestCmdNoAnswer()
                : base(test_cmdNameNoAnswer)
            {
                arguments.Add(new AMP_Parameter("test_val", new AMP_String()));
                response.Add(new AMP_Parameter("ret_val", new AMP_String()));
                errors.Add(typeof(MyAMPError), _b("MY_ERROR"));
                requiresAnswer = false;
            }
        }
        public static TestCmd test_cmd = new TestCmd();
        public static TestCmdNoAnswer test_cmdNoAnswer = new TestCmdNoAnswer();
        public AMP_Msg_Raw test_msg;
        public AMP_Msg_Raw test_msg_no_ask;
        public AMP_Msg_Raw test_msgNoAnswer;
        public Test_AMP_Protocol proto;
        public AMP_Message_Dispatch_Base()
        {
            test_msg = new AMP_Msg_Raw();
            test_msg["_command"] = _b("TestCmd");
            test_msg["_ask"] = _b("25");
            test_msg["test_val"] = _b("Hello World!");

            test_msg_no_ask = new AMP_Msg_Raw();
            test_msg_no_ask["_command"] = _b("TestCmd");
            test_msg_no_ask["test_val"] = _b("Hello World!");

            test_msgNoAnswer = new AMP_Msg_Raw();
            test_msgNoAnswer["_command"] = _b("TestCmdNoAnswer");
            test_msgNoAnswer["_ask"] = _b("25");
            test_msgNoAnswer["test_val"] = _b("Hello World!");
        }

        [SetUp]
        public void Init()
        {
            proto = new Test_AMP_Protocol();
        }

    }

    [TestFixture]
    public class Command_Dispatch_Tests : AMP_Message_Dispatch_Base
    {
        /// <summary>
        /// Test what happens where there is no responder for the given command
        /// </summary>
        [Test]
        public void test_no_responder()
        {
            // dispatch a command for which we have not registered any responder
            var reqMsg = new AMP_Msg_Raw { { "_command", _b("FOO") } };
            proto.processFullMessage(reqMsg);

            // no error should be written, because the request didn't contain an _ask key,
            // so it is not expecting any reply to this command.
            Assert.That(0, Is.EqualTo(proto.messages.Count));

            //ok lets add an _ask key and do it again
            reqMsg.Add("_ask", _b("77"));
            proto.processFullMessage(reqMsg);

            // now we should find an error message has been sent out
            Assert.That(1, Is.EqualTo(proto.messages.Count));

            AMP_Msg_Raw errMsg = proto.messages[0];

            Assert.That(3, Is.EqualTo(errMsg.Count));
            Assert.That(_b("77"), Is.EqualTo(errMsg["_error"]));
            Assert.That(AMP_Command.UNHANDLED_ERROR_CODE, Is.EqualTo(errMsg["_error_code"]));
            Assert.That(errMsg.ContainsKey("_error_description"));
        }

        [Test]
        [ExpectedException(typeof(AMPParameterNotFound))]
        public void test_deserialization_fails()
        {
            // that what happens when the command we receive doesn't match the schema of our local command definition.
            var oddball_msg = new AMP_Msg_Raw
                                  {
                                      {"_ask", _b("1")},
                                      {"_command", _b("TestCmd")},
                                      {"fubar", _b("undefined field")}
                                  };

            // register TestCmd w/ protocol
            proto.RegisterResponder(test_cmd, delegate { return new Object(); });

            // OK, process the command.. this should raise an exception, which this test case expects.
            proto.processFullMessage(oddball_msg);
        }

        [Test]
        public void test_responder_returns_amp_msg()
        {
            // Test a successful command request that results in a call to a responder that returns an AMP_Msg
            var responseMsg = new AMP_Msg { { "ret_val", "Return Value" } };

            var requestMsg = new AMP_Msg();
            Object responderState = "value_before";
            proto.RegisterResponder(test_cmd, delegate(AMP_Msg reqMsg, Object state)
            {
                requestMsg = reqMsg; responderState = state; return responseMsg;
            }, "test_state");

            // process request.. this should result in a successful response being written out
            proto.processFullMessage(test_msg);

            Assert.That(1, Is.EqualTo(requestMsg.Count));
            Assert.That(requestMsg["test_val"], Is.EqualTo("Hello World!"));
            Assert.That(responderState, Is.EqualTo("test_state"));

            Assert.That(1, Is.EqualTo(proto.messages.Count));

            AMP_Msg_Raw respMsgRaw = proto.messages[0];

            Assert.That(2, Is.EqualTo(respMsgRaw.Count));
            Assert.That(_b("25"), Is.EqualTo(respMsgRaw["_answer"]));
            Assert.That(_b("Return Value"), Is.EqualTo(respMsgRaw["ret_val"]));
        }

        [Test]
        public void test_responder_returns_malformed_amp_msg()
        {
            // Test a responder that returns a malformed AMP_Msg
            var responseMsg = new AMP_Msg { { "wrong_key", "foo" } };

            proto.RegisterResponder(test_cmd, delegate { return responseMsg; });

            // process request.. this should result in an error being written out
            // because the responder returns an invalid value
            proto.processFullMessage(test_msg);

            Assert.That(1, Is.EqualTo(proto.messages.Count));

            AMP_Msg_Raw respMsgRaw = proto.messages[0];

            Assert.That(3, Is.EqualTo(respMsgRaw.Count));
            Assert.That(_b("25"), Is.EqualTo(respMsgRaw["_error"]));
            Assert.That(_b("UNKNOWN"), Is.EqualTo(respMsgRaw["_error_code"]));
            Assert.That(respMsgRaw.ContainsKey("_error_description"));
        }

        [Test]
        public void test_responder_returns_delegate_box()
        {
            // Test a successful command request that results in a call to a responder that returns a DelegateBox
            var responseMsg = new AMP_Msg { { "ret_val", "Return Value" } };

            var requestMsg = new AMP_Msg();
            var dBox = new DelegateBox();
            // return delegatebox from responder
            proto.RegisterResponder(test_cmd, delegate(AMP_Msg reqMsg, Object state) { requestMsg = reqMsg; return dBox; });

            // process request.. the proto should attach a callback to the DelegateBox
            proto.processFullMessage(test_msg);

            Assert.That(0, Is.EqualTo(proto.messages.Count));

            Assert.That(1, Is.EqualTo(requestMsg.Count));
            Assert.That(requestMsg["test_val"], Is.EqualTo("Hello World!"));

            // OK, lets fire the DelegateBox callback, which should result in the AMP response being written out
            dBox.callback(responseMsg);

            Assert.That(1, Is.EqualTo(proto.messages.Count));

            AMP_Msg_Raw respMsgRaw = proto.messages[0];

            Assert.That(2, Is.EqualTo(respMsgRaw.Count));
            Assert.That(_b("25"), Is.EqualTo(respMsgRaw["_answer"]));
            Assert.That(_b("Return Value"), Is.EqualTo(respMsgRaw["ret_val"]));
        }

        [Test]
        public void test_responder_returns_delegate_box_which_fires_with_exception()
        {
            var requestMsg = new AMP_Msg();
            var dBox = new DelegateBox();
            // return delegatebox from responder
            proto.RegisterResponder(test_cmd, delegate(AMP_Msg reqMsg, Object state) { requestMsg = reqMsg; return dBox; });

            // process request.. the proto should attach a callback to the DelegateBox
            proto.processFullMessage(test_msg);

            Assert.That(0, Is.EqualTo(proto.messages.Count));

            Assert.That(1, Is.EqualTo(requestMsg.Count));
            Assert.That(requestMsg["test_val"], Is.EqualTo("Hello World!"));

            // OK, lets fire the DelegateBox callback with an exception,
            // which should result in an AMP error being written out
            dBox.callback(new Exception());

            Assert.That(1, Is.EqualTo(proto.messages.Count));

            AMP_Msg_Raw respMsgRaw = proto.messages[0];

            Assert.That(3, Is.EqualTo(respMsgRaw.Count));
            Assert.That(_b("25"), Is.EqualTo(respMsgRaw["_error"]));
            Assert.That(_b("UNKNOWN"), Is.EqualTo(respMsgRaw["_error_code"]));
            Assert.That(respMsgRaw.ContainsKey("_error_description"));
        }

        [Test]
        public void test_no_ask_key()
        {
            // test that we ignore the responder's return value if no _ask key was given
            var responseMsg = new AMP_Msg { { "ret_val", "Return Value" } };

            proto.RegisterResponder(test_cmd, delegate { return responseMsg; });

            // process request.. no response should be written because the peer didn't ask for one
            proto.processFullMessage(test_msg_no_ask);

            Assert.That(0, Is.EqualTo(proto.messages.Count));
        }

        [Test]
        public void test_responder_throws()
        {
            proto.RegisterResponder(test_cmd, delegate { throw new Exception(); });

            proto.processFullMessage(test_msg);

            Assert.That(1, Is.EqualTo(proto.messages.Count));

            AMP_Msg_Raw respMsgRaw = proto.messages[0];

            Assert.That(3, Is.EqualTo(respMsgRaw.Count));
            Assert.That(_b("25"), Is.EqualTo(respMsgRaw["_error"]));
            Assert.That(_b("UNKNOWN"), Is.EqualTo(respMsgRaw["_error_code"]));
            Assert.That(respMsgRaw.ContainsKey("_error_description"));
        }


        [Test]
        public void test_responder_throws_but_no_ask_key_given()
        {
            // test that we don't send an error upon an exception in the responder if no _ask key is given
            proto.RegisterResponder(test_cmd, delegate { throw new Exception(); });
            proto.processFullMessage(test_msg_no_ask);

            // responder threw an exception, but no messages written out. Good.
            Assert.That(0, Is.EqualTo(proto.messages.Count));
        }

        [Test]
        public void test_responder_throws_custom_exception()
        {
            // OK, now test having the responder throw a known exception that has a custom AMP error code

            proto.RegisterResponder(test_cmd, delegate { throw new MyAMPError(); });
            proto.processFullMessage(test_msg);

            Assert.That(1, Is.EqualTo(proto.messages.Count));

            AMP_Msg_Raw respMsgRaw = proto.messages[0];

            Assert.That(3, Is.EqualTo(respMsgRaw.Count));
            Assert.That(_b("25"), Is.EqualTo(respMsgRaw["_error"]));
            Assert.That(_b("MY_ERROR"), Is.EqualTo(respMsgRaw["_error_code"]));
            Assert.That(Encoding.UTF8.GetString(respMsgRaw["_error_description"]).Contains("MyAMPError"));
        }
    }

    public class SimpleCmd : AMP_Command
    {
        public SimpleCmd()
            : base("SimpleCmd")
        {
            arguments.Add(new AMP_Parameter("test_val", new AMP_String()));
            response.Add(new AMP_Parameter("ret_val", new AMP_String()));
        }
    }

    /// <summary>
    /// Test full end-to-end AMP functionality over a loopback socket
    /// </summary>
    [TestFixture]
    public class Loopback_Tests
    {
        private int port = 45343;
        private Socket listenSock;
        private Socket serverSock;
        private Socket clientSock;
        private AMPSocketTransport serverTransport;
        private AMP_Protocol server;
        private AMPSocketTransport clientTransport;
        private AMP_Protocol client;

        [SetUp]
        public void SetUp()
        {
            listenSock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            listenSock.Bind(new IPEndPoint(IPAddress.Parse("127.0.0.1"), port));
            listenSock.Listen(5);

            //Socket serverSock = null;
            var t = new Thread(() => { serverSock = listenSock.Accept(); });
            t.Start();

            clientSock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            clientSock.Connect("localhost", port);
            t.Join();
            //System.Console.WriteLine("serverSock: {0}", serverSock);
            //System.Console.WriteLine("clientSock: {0}", clientSock);
            serverTransport = new AMPSocketTransport(serverSock);
            server = new AMP_Protocol(serverTransport);
            serverTransport.Start();

            clientTransport = new AMPSocketTransport(clientSock);
            client = new AMP_Protocol(clientTransport);
            clientTransport.Start();
        }

        [TearDown]
        public void TearDown()
        {
            clientSock.Close();
            serverSock.Close();
            listenSock.Close();
        }

        [Test]
        public void test_BeginCallRemote_loopback()
        {
            Console.WriteLine("Test thread      : {0}", Thread.CurrentThread.ManagedThreadId);
            Console.WriteLine("test case syn ctx: {0}", SynchronizationContext.Current);
            AMP_Msg reqMsg = null;
            var _ret = new AMP_Msg {{"ret_val", "hello"}};
            server.RegisterResponder(new SimpleCmd(), (msg, state) =>
                                                          {
                                                              reqMsg = msg;
                                                              return _ret;
                                                          });
            var _msg = new AMP_Msg {{"test_val", "value!"}};
            AMP_Msg respMsg = null;
            IAsyncResult _ar = client.BeginCallRemote(new SimpleCmd(), _msg,
                                                      delegate(IAsyncResult ar)
                                                          {
                                                              respMsg = client.EndCallRemote(ar);
                                                          }, "state");
            _ar.AsyncWaitHandle.WaitOne();

            Assert.That("state", Is.EqualTo((string)_ar.AsyncState));

            Assert.That("value!", Is.EqualTo((string)reqMsg["test_val"]));
            Assert.That(1, Is.EqualTo(reqMsg.Count));

            Assert.That("hello", Is.EqualTo((string)respMsg["ret_val"]));
            Assert.That(1, Is.EqualTo(respMsg.Count));
        }

        [Test]
        public void test_CallRemote_loopback()
        {
            AMP_Msg reqMsg = null;
            var _ret = new AMP_Msg {{"ret_val", "hello"}};
            server.RegisterResponder(new SimpleCmd(), (msg, state) =>
                                                          {
                                                              reqMsg = msg;
                                                              return _ret;
                                                          });
            var _msg = new AMP_Msg {{"test_val", "value!"}};
            AMP_Msg respMsg = client.CallRemote(new SimpleCmd(), _msg);

            Assert.That("value!", Is.EqualTo((string)reqMsg["test_val"]));
            Assert.That(1, Is.EqualTo(reqMsg.Count));

            Assert.That("hello", Is.EqualTo((string)respMsg["ret_val"]));
            Assert.That(1, Is.EqualTo(respMsg.Count));
        }

        /// <summary>
        /// Test that registering a responder along with a Control, will actually work,
        /// and actually use that Control to run the responder in the correct thread.
        /// </summary>
        [Test]
        public void testResponderWithStateAndForm()
        {
            // fire up a Form in a new thread and start it running
            Form theForm = null;
            var handle = new ManualResetEvent(false);
            var t = new Thread(new ThreadStart(delegate { theForm = new Form(); handle.Set(); theForm.ShowDialog(); }));
            t.Start();
            handle.WaitOne(); // block until the Thread has started and theForm has been assigned

            var _ret = new AMP_Msg {{"ret_val", "hello"}};
            Object _state = null;
            AMP_Msg _msg = null;
            int _threadID = 0;
            server.RegisterResponder(new SimpleCmd(), (msg, state) =>
                                                          {
                                                              _msg = msg;
                                                              _state = state;
                                                              _threadID = Thread.CurrentThread.ManagedThreadId;
                                                              return _ret;
                                                          },
                                     "state", theForm); // Responder should be invoked on theForm's thread

            AMP_Msg respMsg = client.CallRemote(new SimpleCmd(), new AMP_Msg {{"test_val", "value!"}});

            Assert.That("value!", Is.EqualTo((string)_msg["test_val"]));
            Assert.That(1, Is.EqualTo(_msg.Count));

            Assert.That("hello", Is.EqualTo((string)respMsg["ret_val"]));
            Assert.That(1, Is.EqualTo(respMsg.Count));

            Assert.That("state", Is.EqualTo((string)_state));
            Assert.That(t.ManagedThreadId, Is.EqualTo(_threadID));
            theForm.Invoke(new MethodInvoker(theForm.Close));
        }

        /// <summary>
        /// Ensure that the synchronous CallRemote API works OK when called from a form.
        /// </summary>
        [Test]
        public void testCallRemoteFromForm()
        {
            var _ret = new AMP_Msg {{"ret_val", "hello"}};
            server.RegisterResponder(new SimpleCmd(), (msg, state) => { return _ret; });

            CallRemoteForm theForm = null;
            var t = new Thread(new ThreadStart(delegate
                                                   {
                                                       theForm = new CallRemoteForm(client);
                                                       theForm.ShowDialog();
                                                   }));
            t.Start();
            t.Join();

            Assert.That("hello", Is.EqualTo((string)theForm.respMsg["ret_val"]));
            Assert.That(1, Is.EqualTo(theForm.respMsg.Count));
        }
        // TODO also test registering a responder with an explicit SynchronizationContext
    }

    public class CallRemoteForm: Form
    {
        public AMP_Msg respMsg;
        private AMP_Protocol client;
         public CallRemoteForm(AMP_Protocol client)
         {
             this.client = client;
             Load += CallRemoteForm_OnLoad;
         }
        public void CallRemoteForm_OnLoad(Object sender, EventArgs e)
        {
            respMsg = client.CallRemote(new SimpleCmd(), new AMP_Msg{{"test_val", "value!"}});
            Close();
        }
    }
}
